package eu.cloudapio.deathswap;

import java.io.File;
import java.util.*;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Difficulty;
import org.bukkit.GameMode;
import org.bukkit.GameRule;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.WorldCreator;
import org.bukkit.World.Environment;
import org.bukkit.advancement.Advancement;
import org.bukkit.advancement.AdvancementProgress;
import org.bukkit.entity.Player;

public class DeathSwapGame {
    ArrayList<Player> players;
    WorldCreator wc;
    World world;
    Random randomizer;
    String worldName;

    boolean isPlayingGame = false;

    int minDeathSwapPlayers = 2;
    int minSwapTime = 30;
    int maxSwapTime = 150;
    double swapCancelProbability = 0.15;

    long lastSwapTime;

    public DeathSwapGame()
    {
        randomizer = new Random(System.currentTimeMillis());
        players = new ArrayList<Player>();
        worldName = "death-swap";
        wc = new WorldCreator(worldName);

        getMainWorld().setPVP(false);
    }

    private World getMainWorld()
    {
        return Bukkit.getServer().getWorlds().get(0);
    }

    private boolean isUnspawnableNetherLocation(World world, Location loc)
    {
        if(world.getBlockAt(loc).getType() == Material.AIR) return true;
        if(world.getBlockAt(loc).getType() == Material.BEDROCK) return true;

        if(world.getBlockAt(loc.add(0, 1, 0)).getType() != Material.AIR) return true;
        if(world.getBlockAt(loc.add(0, 2, 0)).getType() != Material.AIR) return true;

        return false;
    }

    // Selects a spawn location not in the air, not in water and not on top of bedrock (nether roof).
    private Location randomSpawnLocation(World world, boolean isNether)
    {
        while(true)
        {
            int x = randomizer.nextInt(-1000000, 1000000);
            int z = randomizer.nextInt(-1000000, 1000000);

            if(Math.abs(x) < 500) continue;
            if(Math.abs(z) < 500) continue;

            int y = 0;

            if(isNether) {
                y = 95;
                while(isUnspawnableNetherLocation(world, new Location(world, x, y, z)))
                {
                    y--;
                    if(y == 0) break;
                }
                if(y == 0) continue;
            }
            else y = world.getHighestBlockYAt(x, z);

            if(world.getBlockAt(x, y, z).getType() == Material.WATER) continue;
            if(world.getBlockAt(x, y, z).getType() == Material.LAVA) continue;

            return new Location(world, x, y, z);
        }

    }

    private void revokeAdvancements(Player p)
    {
        Iterator<Advancement> iterator = Bukkit.getServer().advancementIterator();
        while (iterator.hasNext())
        {
            AdvancementProgress progress = p.getAdvancementProgress(iterator.next());
            for (String criteria : progress.getAwardedCriteria())
                progress.revokeCriteria(criteria);
        }
    }

    private void resetPlayer(Player p)
    {
        p.getInventory().clear();
        p.setHealth(20);
        p.setFoodLevel(20);
        p.setFireTicks(0);
        p.setTotalExperience(0);
        p.setFreezeTicks(0);
        p.setFallDistance(0);
    }

    public int start(boolean isNether)
    {
        if(Bukkit.getServer().getOnlinePlayers().size() < minDeathSwapPlayers)
        {
            return -1;
        }

        if(isPlayingGame)
        {
            return -2;
        }

        players = new ArrayList<Player>(Bukkit.getServer().getOnlinePlayers());

        Bukkit.broadcastMessage(ChatColor.YELLOW + "Starting DeathSwap game...");

        isPlayingGame = true;

        Bukkit.getServer().unloadWorld(world, false);

        deleteWorld(worldName);

        wc.copy(Bukkit.getServer().getWorlds().get(0));
        wc.seed(randomizer.nextLong());
        if(isNether)
        {
            wc.environment(Environment.NETHER);
        }
        world = wc.createWorld();
        world.setKeepSpawnInMemory(false);
        world.setDifficulty(Difficulty.NORMAL);
        world.setPVP(false);
        world.setGameRule(GameRule.DO_IMMEDIATE_RESPAWN, true);

        ArrayList<Location> spawnLocations = new ArrayList<Location>(players.size());
        for(int i = 0; i < players.size(); i++)
        {
            spawnLocations.add(randomSpawnLocation(world, isNether));
        }

        for(int i = 0; i < players.size(); i++)
        {
            Player p = players.get(i);
            revokeAdvancements(p);
            resetPlayer(p);
            p.setGameMode(GameMode.SURVIVAL);
            p.teleport(spawnLocations.get(i));
        }

        int swapTime = getSwapTime();
        int tickTime = swapTime * 20;
        Bukkit.getLogger().log(Level.INFO, String.format("Next swap in %d seconds", swapTime));
        Bukkit.getServer().getScheduler().runTaskLater(DeathSwap.plugin, DeathSwap.notifier, tickTime - 60);
        Bukkit.getServer().getScheduler().runTaskLater(DeathSwap.plugin, DeathSwap.runner, tickTime);
        Bukkit.getServer().getScheduler().runTaskLater(DeathSwap.plugin, DeathSwap.timeNotifier, 20);

        lastSwapTime = world.getFullTime();

        return 0;
    }

    public void teleportToWorldSpawn(Player p)
    {
        Location worldSpawn = getMainWorld().getSpawnLocation();

        p.setGameMode(GameMode.ADVENTURE);
        resetPlayer(p);
        p.teleport(worldSpawn);
    }

    private boolean hasPlayer(Player player)
    {
        for(Player p : players)
        {
            if(p.getUniqueId().equals(player.getUniqueId()))
            {
                return true;
            }
        }

        return false;
    }

    private void checkGameEnd()
    {
        if(players.size() < minDeathSwapPlayers)
        {
            if(players.size() == 0)
            {
                Bukkit.broadcastMessage(ChatColor.translateAlternateColorCodes('&', "&eNo one won &f(&cis this a bug?&f)"));
            }
            else {
                Player winner = players.get(0);
                Bukkit.broadcastMessage(ChatColor.translateAlternateColorCodes('&',
                String.format("&3%s &6won the game!", winner.getName())));
            }

            Bukkit.getScheduler().cancelTasks(DeathSwap.plugin);

            isPlayingGame = false;

            for(Player p : players)
            {
                teleportToWorldSpawn(p);
            }
        }
    }

    public void playerDied(Player player)
    {
        if(!isPlayingGame) return;

        if(hasPlayer(player))
        {
            players.remove(player);
            player.setGameMode(GameMode.ADVENTURE);

            Bukkit.broadcastMessage(ChatColor.translateAlternateColorCodes('&',
                String.format("&3%s &chas died. &9%d players &fremaining!", player.getName(), players.size())));

            checkGameEnd();
        }
    }

    public void playerLeft(Player player)
    {
        if(!isPlayingGame) return;

        if(hasPlayer(player))
        {
            players.remove(player);
            player.setGameMode(GameMode.ADVENTURE);
            teleportToWorldSpawn(player);

            Bukkit.broadcastMessage(ChatColor.translateAlternateColorCodes('&',
                String.format("&3%s &cleft. &9%d players &fremaining!", player.getName(), players.size())));

            checkGameEnd();
        }
    }

    private void deleteWorld(String worldName)
    {
        File container = Bukkit.getServer().getWorldContainer();

        for(File f : container.listFiles())
        {
            if(f.getName().equals(worldName))
            {
                deleteFolder(f);
            }
        }
    }

    private void deleteFolder(File folder)
    {
        if(folder.isDirectory())
        {
            for(File f : folder.listFiles())
            {
                deleteFolder(f);
            }
        }
        folder.delete();
    }

    public void swap()
    {
        lastSwapTime = world.getFullTime();

        if(randomizer.nextDouble() < swapCancelProbability) {
            Bukkit.broadcastMessage(ChatColor.GOLD + "Swap cancelled!");
            return;
        }

        ArrayList<Location> locations = new ArrayList<Location>(players.size() + 1);
        for(Player p : players)
        {
            locations.add(p.getLocation());
        }
        locations.add(locations.get(0));

        for(int i = 0; i < players.size(); i++)
        {
            players.get(i).teleport(locations.get(i+1));
        }
    }

    public int getSwapTime()
    {
        return randomizer.nextInt(minSwapTime, maxSwapTime);
    }

    public long getTicksSinceLastSwap()
    {
        return world.getFullTime() - lastSwapTime;
    }
}
